import logging
import threading
import time

import json

import websocket
from filter_dict import filter_dict

from websocket_communication_server.messagetemplates import commandmessage


class WebSocketClient:
    def __init__(self, name, logger=None, host=None, reconnect=True):
        name = str(name)
        self.message_cmd_functions = {}
        self.message_types = {}
        self.ws = None
        self.time = time.time()
        self._on_open_functions = {}
        self._on_error_functions = {}
        self._on_close_functions = {}
        self._on_message_functions = {}
        if logger is None:
            logger = logging.getLogger("ListenWebsocket_" + name)
        self.logger = logger
        self.ws_thread = None
        self.reconnect = reconnect
        self.host = None
        self.reconnect_time = 1
        self.name = name

        self.add_on_message(name="default", func=self.default_messagevalidator)

        self.add_message_type("cmd", self.default_command_validator)
        self.add_cmd_function(
            "indentify",
            lambda: self.write_to_socket(
                commandmessage(cmd="indentify", sender=self.name, name=self.name)
            ),
        )
        self.add_cmd_function("set_time", self.set_time)
        if host is not None:
            self.connect_to_socket(host)

    def set_time(self, time=None):
        if time is None:
            return
        self.time = time

    def connect_to_socket(self, host):
        if self.ws is not None:
            self.close()
        self.ws = websocket.WebSocketApp(
            host,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
            on_open=self.on_open,
        )
        self.logger.info("conenct to websocket: " + host)
        self.ws_thread = threading.Thread(target=self.ws.run_forever, daemon=True)

        self.ws_thread.start()
        time.sleep(1)
        if self.ws is None:
            self.host = None
            return False
        self.host = host
        self.logger.debug("websocket connected")

    def on_open(self):
        self.logger.debug("websocket opened")
        for n, f in self._on_open_functions.items():
            f()

    def on_message(self, message):
        self.logger.debug("message " + message)
        for n, f in self._on_message_functions.items():
            f(message)

    def on_error(self, ws):
        self.logger.debug("socket error")
        for n, f in self._on_error_functions.items():
            f()

    def on_close(self):
        self.logger.info("websocket closed")
        for n, f in self._on_close_functions.items():
            f()
        self.ws = None
        self.ws_thread = None
        if self.reconnect and self.host is not None:
            time.sleep(self.reconnect_time)
            self.connect_to_socket(self.host)

    def add_on_open(self, name, func, overwrite=False):
        if name in self._on_open_functions and not overwrite:
            self.logger.error("on_open_functions " + name + " already defined!")
            return
        self._on_open_functions[name] = func

    def add_on_error(self, name, func, overwrite=False):
        if name in self._on_error_functions and not overwrite:
            self.logger.error("on_error_functions " + name + " already defined!")
            return
        self._on_error_functions[name] = func

    def add_on_close(self, name, func, overwrite=False):
        if name in self._on_close_functions and not overwrite:
            self.logger.error("on_close_function " + name + " already defined!")
            return
        self._on_close_functions[name] = func

    def add_on_message(self, name, func, overwrite=False):
        if name in self._on_message_functions and not overwrite:
            self.logger.error("on_message_function " + name + " already defined!")
            return
        self._on_message_functions[name] = func

    def write_to_socket(self, msg):
        if self.ws is not None:
            if self.ws_thread is not None:
                try:
                    self.ws.send(msg)
                except Exception as e:
                    self.logger.exception(e)

    def close(self):
        self.reconnect = False
        if self.ws is not None:
            self.ws.close()
            if self.ws_thread is not None:
                try:
                    self.ws_thread.join()
                except:
                    pass
                self.ws_thread = None
            self.ws = None

    def default_command_validator(self, data):
        cmd_data = data["data"]
        if cmd_data["cmd"] in self.message_cmd_functions:
            try:
                self.message_cmd_functions[cmd_data["cmd"]](
                    **filter_dict(
                        cmd_data["kwargs"], self.message_cmd_functions[cmd_data["cmd"]]
                    )
                )
            except Exception as e:
                self.logger.exception(e)
        else:
            self.logger.error("unknown command: " + cmd_data["cmd"])

    def default_messagevalidator(self, msg):
        try:
            jmsg = json.loads(msg)
            if jmsg["type"] in self.message_types:
                self.message_types[jmsg["type"]](jmsg)
            else:
                self.logger.error("unknown message type " + msg)
        except:
            self.logger.exception(Exception)

    def add_cmd_function(self, name, func, overwrite=False):
        if name in self.message_cmd_functions and not overwrite:
            self.logger.error("function " + name + " already defined!")
            return
        self.message_cmd_functions[name] = func

    def add_message_type(self, name, func, overwrite=False):
        if name in self.message_types and not overwrite:
            self.logger.error("message type " + name + " already defined!")
            return
        self.message_types[name] = func



import asyncio

import json

from websocket_communication_server.messagetemplates import commandmessage


class Connection:
    def __init__(self, ws, server):
        self.server = server
        self.server.logger.debug("create connection")
        self.ws = ws
        self.identified = False
        self.name = "unknown"
        self.loop = asyncio.get_event_loop()

    def ask_for_identification(self):
        self.server.logger.debug("ask for identification")
        self.sendMsg(commandmessage(sender="server", cmd="indentify"))
        self.sendMsg(
            commandmessage(sender="server", cmd="set_time", time=self.server.t0)
        )

    def sendMsg(self, msg):
        coro = self.ws.send(msg)
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)

    async def recive(self):
        try:
            async for message in self.ws:
                self.validate_message(message)
                if not self.identified:
                    self.ask_for_identification()
        except asyncio.IncompleteReadError:
            pass

    def validate_message(self, msg):
        self.server.logger.debug(self.name + " recived message: " + msg)
        try:
            jmsg = json.loads(msg)
            if not self.identified:
                if "cmd" not in jmsg["data"]:
                    return
                if jmsg["data"]["cmd"] != "indentify":
                    return
            if "server" in jmsg["target"]:
                if jmsg["type"] == "cmd":
                    self.run_command(jmsg)
                elif jmsg["type"] == "data":
                    pass
                else:
                    self.server.logger.error("unknown message type " + msg)
            return self.server.send_to_names(jmsg["target"], msg)
        except:
            self.server.logger.exception(Exception)
            pass

    def run_command(self, data):
        cmd_data = data["data"]
        if cmd_data["cmd"] == "indentify":
            self.identify(data)
        else:
            self.server.logger.error("unknown command: " + str(cmd_data["cmd"]))

    def identify(self, data):
        cmd_data = data["data"]
        if "name" in cmd_data["kwargs"]:
            self.name = cmd_data["kwargs"]["name"]
            self.identified = True



import setuptools

setup = dict(
    name="WsComSrv",
    version="0.1",
    author="Julian Kimmig",
    author_email="julian-kimmig@gmx.net",
    description="A basic websocket communication server",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/JulianKimmig/websocket_communication_server",
    packages=setuptools.find_packages(),
    install_requires=['FilterDict','websocket-client','websockets'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)

if __name__ == "__main__":
    setuptools.setup(**setup)
